import 'dart:math';

import 'package:uuid/uuid.dart';

/// iOS设备的产品名称
enum ProductName {
  IPHONE_7,
  IPHONE_7_PLUS,
  IPHONE_8,
  IPHONE_8_PLUS,
  IPHONE_X,
  IPHONE_XR,
  IPHONE_XS,
  IPHONE_XS_MAX,
  // ADD_ONE_MARKER
}

/// iOS 设备类型名称
///
/// 参考 https://www.theiphonewiki.com/wiki/Models
const Map<ProductName, String> DeviceTypeTable = {
  ProductName.IPHONE_7: "iPhone9,1",
  ProductName.IPHONE_7_PLUS: "iPhone9,2",
  ProductName.IPHONE_8: "iPhone10,1",
  ProductName.IPHONE_8_PLUS: "iPhone10,2",
  ProductName.IPHONE_X: "iPhone10,3",
  ProductName.IPHONE_XR: "iPhone10,8",
  ProductName.IPHONE_XS: "iPhone11,2",
  ProductName.IPHONE_XS_MAX: "iPhone11,6"
  // ADD_ONE_MARKER
};

enum SystemVersion {
  IOS_10_0,
  IOS_10_1,
  IOS_10_2,
  IOS_10_3,
  IOS_11_0,
  IOS_11_1,
  IOS_11_2,
  IOS_11_3,
  IOS_11_4,
  IOS_12_0,
  IOS_12_1,
  IOS_12_2,
  IOS_12_3,
  // ADD_ONE_MARKER
  IOS_LASTEST
}

/// iOS系统固件版本
/// 记录每个系统版本下最新的固件版本
///
/// 参考 http://jailbreak.25pp.com/gujian/
const Map<SystemVersion, String> FirmwareVersionTable = {
  SystemVersion.IOS_10_0: "10.0.3",
  SystemVersion.IOS_10_1: "10.1.1",
  SystemVersion.IOS_10_2: "10.2.1",
  SystemVersion.IOS_10_3: "10.3.3",
  SystemVersion.IOS_11_0: "11.0.3",
  SystemVersion.IOS_11_1: "11.1.2",
  SystemVersion.IOS_11_2: "11.2.6",
  SystemVersion.IOS_11_3: "11.3.1",
  SystemVersion.IOS_11_4: "11.4.1",
  SystemVersion.IOS_12_0: "12.0.1",
  SystemVersion.IOS_12_1: "12.1.4",
  SystemVersion.IOS_12_2: "12.2",
  SystemVersion.IOS_12_3: "12.3.1",
  // ADD_ONE_MARKER
};

/// iOS各个设备支持的最高系统版本
///
/// 系统升级只能升级到当前设备支持的最高版本
const Map<ProductName, SystemVersion> SupremeSupports = {
  ProductName.IPHONE_7: SystemVersion.IOS_LASTEST,
  ProductName.IPHONE_7_PLUS: SystemVersion.IOS_LASTEST,
  ProductName.IPHONE_8: SystemVersion.IOS_LASTEST,
  ProductName.IPHONE_8_PLUS: SystemVersion.IOS_LASTEST,
  ProductName.IPHONE_X: SystemVersion.IOS_LASTEST,
  ProductName.IPHONE_XR: SystemVersion.IOS_LASTEST,
  ProductName.IPHONE_XS: SystemVersion.IOS_LASTEST,
  ProductName.IPHONE_XS_MAX: SystemVersion.IOS_LASTEST,
  // ADD_ONE_MARKER
};

class Display {
  final int width;
  final int height;
  final int ppi;
  final double inch;

  /// 设备屏幕缩放率
  final double scale;

  /// 设备屏幕物理缩放率
  final double nativeScale;

  /// 保持较大值在前,较小值在后
  const Display(this.width, this.height, this.ppi, this.inch, this.scale,
      this.nativeScale);

  String resolutionToString({bool reverse = false}) =>
      reverse ? "${height}x$width" : "${width}x$height";

  static const Map<ProductName, Display> Table = {
    ProductName.IPHONE_7: Display(1334, 750, 326, 4.7, 3.00, 3.00),
    ProductName.IPHONE_7_PLUS: Display(1920, 1080, 401, 5.5, 3.00, 3.00),
    ProductName.IPHONE_8: Display(1334, 750, 326, 4.7, 3.00, 3.00),
    ProductName.IPHONE_8_PLUS: Display(1920, 1080, 401, 5.5, 3.00, 3.00),
    ProductName.IPHONE_X: Display(2436, 1125, 458, 5.8, 3.00, 3.00),
    ProductName.IPHONE_XR: Display(1792, 828, 326, 6.1, 3.00, 3.00),
    ProductName.IPHONE_XS: Display(2436, 1125, 458, 5.8, 3.00, 3.00),
    ProductName.IPHONE_XS_MAX: Display(2688, 1242, 458, 6.5, 3.00, 3.00),
    // ADD_ONE_MARKER
  };
}

class QuarkDevice {
  static ProductName randomProductName() {
    final values = ProductName.values;
    return values[Random().nextInt(values.length)];
  }

  /// 已录入的最新系统版本
  static SystemVersion get lastestSystemVersion =>
      SystemVersion.values[SystemVersion.IOS_LASTEST.index - 1];

  /// 返回设备支持的最高系统版本
  static SystemVersion querySupermeVersion([ProductName? productName]) {
    if (productName == null) return lastestSystemVersion;

    SystemVersion? supermeVersion = SupremeSupports[productName];
    if (supermeVersion == SystemVersion.IOS_LASTEST)
      return lastestSystemVersion;
    return supermeVersion ?? SystemVersion.IOS_10_0;
  }

  /// 返回设备支持的系统版本
  static SystemVersion fixupSystemVersion(SystemVersion systemVersion,
      {ProductName? productName, int upgradeOSTime = 0}) {
    final SystemVersion supermeVersion = querySupermeVersion(productName);
    final values = SystemVersion.values;

    final s0 = values.indexOf(systemVersion);
    final s1 = values.indexOf(supermeVersion);

    if (s0 > s1) {
      systemVersion = supermeVersion;
    } else if (s0 < s1) {
      if (upgradeOSTime != 0 &&
          DateTime.now().millisecondsSinceEpoch >= upgradeOSTime) {
        systemVersion = supermeVersion;
      }
    }
    return systemVersion;
  }

  final ProductName productName;
  final SystemVersion systemVersion;
  final String identifier;
  final int upgradeOSTime;

  /// 外部请勿直接使用
  const QuarkDevice(this.productName, this.systemVersion, this.identifier,
      this.upgradeOSTime);

  /// 克隆一台新的设备
  ///
  /// 仅用于继承
  QuarkDevice.clone(QuarkDevice device)
      : this(device.productName, device.systemVersion, device.identifier,
            device.upgradeOSTime);

  /// 从设备库中随机选择一种
  ///
  /// 系统版本初始化为该设备目前已知的最高版本
  factory QuarkDevice.random() => QuarkDevice.fromJson({});

  /// 从JSON中实例化出合法的设备
  factory QuarkDevice.fromJson(Map<String, dynamic>? json) {
    json ??= {};
    String deviceType = json["device_type"] ?? "";
    ProductName productName = DeviceTypeTable.keys.singleWhere(
        (k) => DeviceTypeTable[deviceType] != null,
        orElse: randomProductName);

    String firmwareVersion = json["firmware_version"] ?? "";
    SystemVersion systemVersion = FirmwareVersionTable.keys.singleWhere(
        (k) => FirmwareVersionTable[firmwareVersion] != null,
        orElse: () => SystemVersion.IOS_LASTEST);

    int upgradeOSTime = json["os_upgrade_ts"] ?? 0;

    systemVersion = fixupSystemVersion(systemVersion,
        productName: productName, upgradeOSTime: upgradeOSTime);

    String deviceId = json["device_id"] ?? Uuid().v4().toUpperCase();

    // 升级系统检查, 延迟升级2~3天
    if (systemVersion == querySupermeVersion(productName)) {
      upgradeOSTime = 0;
    } else if (upgradeOSTime == 0) {
      upgradeOSTime = DateTime.now().millisecondsSinceEpoch +
          2 * 24 * 60 * 60 +
          Random().nextInt(1 * 24 * 60 * 60);
    }
    return QuarkDevice(productName, systemVersion, deviceId, upgradeOSTime);
  }

  Map<String, dynamic> toJson() {
    return {
      "device_type": deviceType,
      "firmware_version": firmwareVersion,
      "os_upgrade_ts": upgradeOSTime,
      "device_id": identifier,
    };
  }

  String get deviceType {
    return DeviceTypeTable[productName] ?? "iPhone9,1";
  }

  String get firmwareVersion {
    return FirmwareVersionTable[systemVersion] ?? "10.0.3";
  }

  Display get display {
    return Display.Table[productName] ?? Display(1334, 750, 326, 4.7, 3.00, 3.00);
  }
}
