
import 'dart:convert';
import 'dart:io';
import 'dart:math';
import 'dart:typed_data';

import 'package:igapi/src/errors.dart';
import 'package:uuid/uuid.dart';

import 'config.dart';
import 'network.dart';
import 'models.dart';
import 'actions.dart';
import 'quark_ig.dart';

class IGAPIGateway {
  IGAccount? _account;

  String? get _deviceId => _account?.device.identifier;
  String? get _userId => _account?.profile?.userId;
  Future<String> get _csrftoken async {
    final cookies =
        await _account?.cookieJar.loadForRequest(Uri.parse(IGAPIConfig.API_V1)) ??
            [];
    for (Cookie cookie in cookies) {
      if (cookie.name == "csrftoken") {
        return cookie.value;
      }
    }
    return "missing";
  }

  NetResolve? netResolve;
  NetReject?  netReject;

  APINet get _apiNet {
    final apiNet = _account?.apiNet;
    if (apiNet == null) {
      throw IGException(IGExceptionType.LoginRequired, "login required");
    }
    apiNet.netResolve = netResolve;
    apiNet.netReject = netReject;
    return apiNet;
  }

  static String _makeUUID({bool toUpperCase = true, bool separate = true}) {
    String uuid = Uuid().v4();
    if (toUpperCase == true) uuid = uuid.toUpperCase();
    if (separate != true) uuid = uuid.replaceAll("-", "");
    return uuid;
  }

  IGProfile? get profile => _account?.profile;
  bool get loggedIn => !(profile?.expired ?? true);

  IGAPIGateway._getPersist(IGAccount? account) : _account = account;

  // 登录专用
  IGAPIGateway.forLogin() : this._getPersist(null);

  /// 获取持久化账户
  /// 未找到则返回[null]
  ///
  /// 可能是一个登录失效的账户,注意检查
  static Future<IGAPIGateway> getPersist(IGProfile profile) async {
    final account = await IGAccount.getPersist(profile.userId);
    return Future.value(
        account != null ? IGAPIGateway._getPersist(account) : null);
  }

  /// 按由新到老的登录时间排序, 登录失效的也会被纳入到
  static Future<List<IGProfile>> getAllProfiles() {
    return IGProfile.getAllPersistProfiles();
  }

  /// 不再使用时请下线
  void offline() {
    _account?.offline();
    _account = null;
  }

  // APIs

  Future<SXTuple1<IGLoggedInUser>> login(
      String username, String password) async {
    _account = await IGAccount.forLogin(username);

    final body = {
      "reg_login": "0",
      "username": username,
      "password": password,
      "device_id": _deviceId,
      "phone_id": _deviceId,
      "login_attempt_count": (_account?.state.loginAttemptCount ?? 0) + 1
    };
    return (await _apiNet.postSignedJson("/accounts/login", body,
            action: IGAPIAction.AuthLogin))
        .tuple1();
  }

  Future<SXTuple1<IGLoggedInUser>> loginFB(
      String fbuid, String accessToken) async {
    _account = await IGAccount.forLogin(fbuid, facebook: true);

    final body = {
      "dryrun": "true",
      "allow_contacts_sync": "false",
      "phone_id": "",
      "waterfall_id": _makeUUID(toUpperCase: false),
      "device_id": _deviceId,
      "fb_access_token": accessToken
    };
    return (await _apiNet.postSignedJson("/fb/facebook_signup", body,
            action: IGAPIAction.AuthFBLogin))
        .tuple1();
  }

  void _checkTwoFactorLock() {
    if (_account?.state.twoFactorLock == null) {
      throw IGException(
          IGExceptionType.InvalidParameters, "two factor lock not found");
    }
  }

  void _checkChallengelock() {
    if (_account?.state.challengeLock == null) {
      throw IGException(
          IGExceptionType.InvalidParameters, "challenge lock not found");
    }
  }

  void _checkPagination(IGPagination? pagi) {
    if (pagi is! IGPagination || pagi.end) {
      throw IGException(
          IGExceptionType.InvalidParameters, "invalid pagi", pagi);
    }
  }

  Future<SXTuple1<IGLoggedInUser>> loginTFA(String code) async {
    _checkTwoFactorLock();

    final body = {
      "device_id": _deviceId,
      "username": _account?.state.twoFactorLock?.username,
      "two_factor_identifier": _account?.state.twoFactorLock?.tfaId,
      "verification_code": code
    };
    return (await _apiNet.postSignedJson("/accounts/two_factor_login", body,
            action: IGAPIAction.AuthTFLogin))
        .tuple1();
  }

  /// 仅用于重新发送短信验证码时使用
  Future<void> sendSmsForTFA() async {
    _checkTwoFactorLock();

    final waterfallId = _makeUUID(toUpperCase: false);
    final body = {
      "device_id": _deviceId,
      "username": _account?.state.twoFactorLock?.username,
      "two_factor_identifier": _account?.state.twoFactorLock?.tfaId,
      "waterfall_id": waterfallId
    };
    return (await _apiNet.postSignedJson(
            "/accounts/send_two_factor_login_sms", body,
            action: IGAPIAction.AuthTFSmsSend))
        .tuple0();
  }

  Future<SXTuple1<IGChallengeSelection>> fetchChallenge() async {
    _checkChallengelock();

    final apiPath = _account?.state.challengeLock?.apiPath;
    Map<String, dynamic> queryItems = {"device_id": _deviceId};
    return (await _apiNet.get(apiPath,
            queryItems: queryItems,
            directory: false,
            action: IGAPIAction.AuthChallengeFetch))
        .tuple1();
  }

  /// 选择认证
  Future<SXTuple1<IGChallengeVerification>> selectChallenge(
      IGChallengeChoice choice,
      {IGChallengeVerification? replay}) async {
    _checkChallengelock();

    int choiceInt = 0;
    switch (choice) {
      case IGChallengeChoice.Phone:
        choiceInt = 0;
        break;
      case IGChallengeChoice.Email:
        choiceInt = 1;
        break;
      case IGChallengeChoice.ThisWasMe:
        choiceInt = 0;
        break;
      case IGChallengeChoice.Uncaught:
        assert(false);
        choiceInt = 0;
        break;
    }

    var apiPath = _account?.state.challengeLock?.apiPath;
    final body = {
      "_csrftoken": _csrftoken,
      "_uuid": _deviceId,
      "device_id": _deviceId,
      "choice": "$choiceInt",
      // "_uid": "__logged_out_998288.664296"
    };
    bool directory = false;

    if (replay != null) {
      apiPath = "/challenge/replay/${replay.userId}/${replay.nonce}";
      body["phone_number"] = null;
      directory = true;
    }

    return (await _apiNet.postSignedJson(apiPath, body,
            directory: directory, action: IGAPIAction.AuthChallengeSelect))
        .tuple1();
  }

  /// 返回值可能为[null].
  ///
  /// Case 1: 开启TFA验证后
  ///
  /// Case 2: 验证ThisWasMe
  Future<SXTuple1<IGLoggedInUser>> verifyChallenge(String code) async {
    _checkChallengelock();

    final apiPath = _account?.state.challengeLock?.apiPath;
    final body = {"device_id": _deviceId, "security_code": code};
    return (await _apiNet.postSignedJson(apiPath, body,
            directory: false, action: IGAPIAction.AuthChallengeVerify))
        .tuple1();
  }

  Future<SXTuple1<IGAccountUser>> fetchProfile() async {
    return (await _apiNet.get("/accounts/current_user",
            queryItems: {"edit": true},
            action: IGAPIAction.AccountProfileFetch))
        .tuple1();
  }

  Future<SXTuple1<IGUserLite>> switchPrivacy(bool? public) async {
    public = public ?? false;
    final body = {
      "_csrftoken": _csrftoken,
      "_uuid": _deviceId,
    };
    return (await _apiNet.postSignedJson(
            "/accounts/${public ? "set_public" : "set_private"}", body,
            action: IGAPIAction.AccountPrivacySwitch))
        .tuple1();
  }

  Future<SXTuple1<IGCommenterControl>> fetchCommenterControl() async {
    int ret = (await _apiNet.get("/accounts/get_comment_audience_control_types",
            action: IGAPIAction.AccountCommenterControlFetch))
        .tryGet(0);
    return Future.value(SXTuple1.value(IGCommenterControl.values[ret]));
  }

  Future<void> switchCommenterControl(IGCommenterControl control) async {
    String audienceControl;
    switch (control) {
      case IGCommenterControl.Uncaught:
      case IGCommenterControl.Any:
        audienceControl = "any";
        break;
      case IGCommenterControl.FollowingAndFollower:
        audienceControl = "following_and_follower";
        break;
      case IGCommenterControl.Following:
        audienceControl = "following";
        break;
      case IGCommenterControl.Follower:
        audienceControl = "follower";
        break;
    }
    final body = {
      "_csrftoken": _csrftoken,
      "_uuid": _deviceId,
      "_uid": _userId,
      "audience_control": audienceControl
    };
    return (await _apiNet.postSignedJson(
            "/accounts/set_comment_audience_control_type", body,
            action: IGAPIAction.AccountCommenterControlSwitch))
        .tuple0();
  }

  Future<SXTuple1<IGUserGraphStatus>> followUser(String userId) async {
    final body = {
      "_csrftoken": _csrftoken,
      "_uuid": _deviceId,
      "_uid": _userId,
      "user_id": userId
    };
    return (await _apiNet.postSignedJson("/friendships/create/$userId", body,
            action: IGAPIAction.UserGraphFollow))
        .tuple1();
  }

  Future<SXTuple1<IGUserGraphStatus>> unfollowUser(String userId) async {
    final body = {
      "_csrftoken": _csrftoken,
      "_uuid": _deviceId,
      "_uid": _userId,
      "user_id": userId
    };

    return (await _apiNet.postSignedJson("/friendships/destroy/$userId", body,
            action: IGAPIAction.UserGraphUnfollow))
        .tuple1();
  }

  Future<SXTuple1<IGUserGraphStatus>> fetchUserGraph(String userId) async {
    return (await _apiNet.get("/friendships/show/$userId",
            action: IGAPIAction.UserGraphFetch))
        .tuple1();
  }

  Future<SXTuple1<Map<String, IGUserGraphStatus>>> fetchUserGraphBatch(
      List<String> userIds) async {
    final body = {
      "_csrftoken": _csrftoken,
      "_uuid": _deviceId,
      "_uid": _userId,
      "include_reel_info": "0",
      "user_ids": userIds.join(",")
    };
    return (await _apiNet.postSignedJson("/friendships/show_many", body,
            action: IGAPIAction.UserGraphBatchFetch))
        .tuple1();
  }

  Future<SXTuple2<List<IGUserLite>, IGPagination>> fetchFollowers(
      String userId, IGPagination pagi) async {
    _checkPagination(pagi);

    Map<String, dynamic> queryItems = {};

    if (pagi.turning) {
      queryItems["max_id"] = pagi.maxId;
    } else {
      queryItems["rank_token"] = "${_userId}_${_makeUUID()}";
      queryItems["rank_mutual"] = "0";
    }

    return (await _apiNet.get("/friendships/$userId/followers",
            queryItems: queryItems, action: IGAPIAction.FollowerFetch))
        .tuple2();
  }

  Future<SXTuple2<List<IGUserLite>, IGPagination>> fetchFollowees(
      String userId, IGPagination pagi) async {
    _checkPagination(pagi);

    Map<String, dynamic> queryItems = {};

    if (pagi.turning) {
      queryItems["max_id"] = pagi.maxId;
    } else {
      queryItems["rank_token"] = "${_userId}_${_makeUUID()}";
      queryItems["rank_mutual"] = "0";
    }

    return (await _apiNet.get("/friendships/$userId/following",
            queryItems: queryItems, action: IGAPIAction.FolloweeFetch))
        .tuple2();
  }

  Future<SXTuple1<IGUser>> fetchUserInfo(String userId) async {
    return (await _apiNet.get("/users/$userId/info",
            queryItems: {"device_id": _deviceId},
            action: IGAPIAction.UserInfoFetch))
        .tuple1();
  }

  Future<SXTuple1<IGUser>> fetchUserInfoByName(String username) async {
    return (await _apiNet.get("/users/$username/usernameinfo",
            queryItems: {
              "device_id": _deviceId,
              "from_module": "single_feed_profile"
            },
            action: IGAPIAction.UserInfoFetchByName))
        .tuple1();
  }

  Future<SXTuple1<IGMedia>> fetchMediaInfo(String mediaId) async {
    return (await _apiNet.get("/media/$mediaId/info",
            action: IGAPIAction.MediaInfoFetch))
        .tuple1();
  }

  Future<SXTuple1<Map<String, IGUser>>> fetchMediaInfoBatch(
      List<String> mediaIds) async {
    Map<String, dynamic> queryItems = {"media_ids": mediaIds.join(",")};
    return (await _apiNet.get("/media/infos",
            queryItems: queryItems, action: IGAPIAction.MediaInfoBatchFetch))
        .tuple1();
  }

  Future<SXTuple1<List<IGUserLite>>> fetchMediaLikers(String mediaId) async {
    final queryItems = {
      "rank_mutual": "0",
      "query": "",
      "rank_token": _userId ?? "" + '_' + _makeUUID()
    };

    return (await _apiNet.get("/media/$mediaId/likers", queryItems: queryItems, action: IGAPIAction.MediaLikerFetch)).tuple1();
  }

  Future<SXTuple3<List<IGComment>, IGMediaCaption, IGPagination>> fetchMediaComments(String mediaId, IGPagination pagi) async {
    _checkPagination(pagi);
    
    final queryItems = {
      "can_support_threading": "true",
      "max_id": pagi.maxId
    };

    return (await _apiNet.get("/media/$mediaId/comments", queryItems: queryItems, action: IGAPIAction.MediaCommentFetch)).tuple3();
  }

  Future<void> likeMedia(String mediaId) async {
    final body = {
      "_csrftoken": _csrftoken,
      "_uuid": _deviceId,
      "_uid": _userId,
      "module_name": "photo_view",
      "media_id": mediaId,
      "user_id": mediaId.split('_')[1],
    };
    return (await _apiNet.postSignedJson("/media/$mediaId/like", body,
            queryItems: {"d": 0}, action: IGAPIAction.MediaLike))
        .tuple0();
  }

  Future<void> unlikeMedia(String mediaId) async {
    final body = {
      "_csrftoken": _csrftoken,
      "_uuid": _deviceId,
      "_uid": _userId,
      "module_name": "photo_view",
      "media_id": mediaId,
      "user_id": mediaId.split('_')[1]
    };
    return (await _apiNet.postSignedJson("/media/$mediaId/unlike", body,
            action: IGAPIAction.MediaUnlike))
        .tuple0();
  }

  Future<void> saveMedia(String mediaId) async {
    final body = {
      "_csrftoken": _csrftoken,
      "_uuid": _deviceId,
      "_uid": _userId,
      "module_name": "single_feed_profile",
    };
    return (await _apiNet.postSignedJson("/media/$mediaId/save", body,
            action: IGAPIAction.MediaSave))
        .tuple0();
  }

  Future<void> unsaveMedia(String mediaId) async {
    final body = {
      "_csrftoken": _csrftoken,
      "_uuid": _deviceId,
      "_uid": _userId,
      "module_name": "single_feed_profile",
    };
    return (await _apiNet.postSignedJson("/media/$mediaId/unsave", body,
            action: IGAPIAction.MediaUnsave))
        .tuple0();
  }

  Future<void> viewMedia(IGMedia media) async {
    final apiPath = "https://graph.instagram.com/logging_client_events";
    final body = {
      "message": json.encode(_makeMediaViewMessage(this, media)),
      "compressed": "0",
      "access_token": IGAPIConfig.FACEBOOK_ID + "|" + IGAPIConfig.LOG_SECRET_KEY,
      "format": "json"
    };
    return (await _apiNet.postForm(apiPath, body, directory: false,
            action: IGAPIAction.MediaPostedView))
        .tuple0();
  }

  Future<void> viewStory(IGMedia story) async {
    final reelId = "${story.mediaId}_${story.owner.userId}";
    int aTime = story.takenAt.millisecondsSinceEpoch ~/ 1000;
    int bTime = (DateTime.now().millisecondsSinceEpoch / 1000 -
            Random().nextDouble() * 10)
        .toInt();

    final reelTime = "${aTime}_$bTime";
    final apiPath = "${IGAPIConfig.API_V2}/media/seen";
    final query = {"reel": "1", "live_vod": "0"};
    final body = {
      "_csrftoken": _csrftoken,
      "_uuid": _deviceId,
      "_uid": _userId,
      "live_vois": {},
      "reels": {
        reelId: [reelTime]
      }
    };
    return (await _apiNet.postSignedJson(apiPath, body,
            queryItems: query, action: IGAPIAction.MediaStoryView))
        .tuple0();
  }

  Future<SXTuple1<IGComment>> commentMedia(String mediaId, String? text) async {
    final token = _makeUUID(toUpperCase: false);
    final body = {
      "_csrftoken": _csrftoken,
      "_uuid": _deviceId,
      "_uid": _userId,
      "comment_text": text,
      "idempotence_token": token,
      "containermodule": 'comments_v2_single_feed_profile',
    };
    return (await _apiNet.postSignedJson("/media/$mediaId/comment", body, action: IGAPIAction.MediaComment)).tuple1();
  }

  Future<void> switchMediaCommentable(String? mediaId, bool? commentable) async {
    commentable = commentable ?? false;
    final body = {
      "_csrftoken": _csrftoken,
      "_uuid": _deviceId,
    };
    return (await _apiNet.postSignedJson("/media/$mediaId/${commentable?'enable_comments':'disable_comments'}", body, action: IGAPIAction.MediaCommentableSwitch)).tuple0();
  }

  Future<SXTuple2<List<IGMedia>, IGPagination>> fetchPostedFeed(
      String userId, IGPagination pagi) async {
    _checkPagination(pagi);

    Map<String, dynamic> queryItems = {};

    if (pagi.turning) {
      queryItems["max_id"] = pagi.maxId;
    }
    return (await _apiNet.get("/feed/user/$userId",
            queryItems: queryItems, action: IGAPIAction.MediaPostedFeedFetch))
        .tuple2();
  }

  Future<SXTuple2<List<IGMedia>, IGPagination>> fetchPostedFeedByName(
      String username, IGPagination pagi) async {
    _checkPagination(pagi);

    Map<String, dynamic> queryItems = {};

    if (pagi.turning) {
      queryItems["max_id"] = pagi.maxId;
    }
    return (await _apiNet.get("/feed/user/$username/username",
            queryItems: queryItems,
            action: IGAPIAction.MediaPostedFeedFetchByName))
        .tuple2();
  }

  Future<SXTuple2<List<IGMedia>, IGPagination>> fetchLikedFeed(
      IGPagination pagi) async {
    _checkPagination(pagi);

    Map<String, dynamic> queryItems = {};

    if (pagi.turning) {
      queryItems["max_id"] = pagi.maxId;
    }
    return (await _apiNet.get("/feed/liked",
            queryItems: queryItems, action: IGAPIAction.MediaLikedFeedFetch))
        .tuple2();
  }

  Future<SXTuple2<List<IGMedia>, IGPagination>> fetchSavedFeed(
      IGPagination pagi) async {
    _checkPagination(pagi);

    Map<String, dynamic> queryItems = {};

    if (pagi.turning) {
      queryItems["max_id"] = pagi.maxId;
    }
    return (await _apiNet.get("/feed/saved",
            queryItems: queryItems, action: IGAPIAction.MediaSavedFeedFetch))
        .tuple2();
  }

  Future<SXTuple2<List<IGMedia>, IGPagination>> fetchUsertagFeed(
      String userId, IGPagination pagi) async {
    _checkPagination(pagi);

    Map<String, dynamic> queryItems = {};

    if (pagi.turning) {
      queryItems["max_id"] = pagi.maxId;
    }
    return (await _apiNet.get("/usertags/$userId/feed",
            queryItems: queryItems, action: IGAPIAction.MediaUsertagFeedFetch))
        .tuple2();
  }

  Future<SXTuple2<List<IGMedia>, IGPagination>> fetchHashtagFeed(
      String tagName, IGPagination pagi) async {
    _checkPagination(pagi);

    Map<String, dynamic> queryItems = {};

    if (pagi.turning) {
      queryItems["max_id"] = pagi.maxId;
    }
    return (await _apiNet.get("/feed/tag/$tagName",
            queryItems: queryItems, action: IGAPIAction.MediaHashtagFeedFetch))
        .tuple2();
  }

  Future<SXTuple1<List<IGMedia>>> fetchStoryFeed(String userId) async {
    return (await _apiNet.get("/feed/user/$userId/story",
            action: IGAPIAction.MediaStoryFeedFetch))
        .tuple1();
  }

  Future<SXTuple2<List<IGMediaGrid>, IGPagination>> fetchLocationFeed(
      String locationId, IGPagination pagi,
      {bool viewAsRecent = false}) async {
    _checkPagination(pagi);

    final sessionId = "${_userId}_${_makeUUID()}";

    Map<String, dynamic> queryItems = {
      "session_id": sessionId,
      "surface": "grid",
      "tab": viewAsRecent == true ? "recent" : "ranked"
    };

    Map<String, String> body = {
      "_csrftoken": await _csrftoken,
      "_uuid": _deviceId ?? "",
      "session_id": sessionId,
      "_surface": "grid",
      "tab": viewAsRecent == true ? "recent" : "ranked"
    };

    if (pagi.turning) {
      queryItems["max_id"] = pagi.maxId;
      body["max_id"] = pagi.maxId!;
    }

    final page = pagi.tryGet("page");
    if (page != null) {
      queryItems["page"] = page;
      body["page"] = page;
    }
    return (await _apiNet.postForm("/locations/$locationId/sections", body,
            action: IGAPIAction.MediaLocationFeedFetch))
        .tuple2();
  }

  Future<SXTuple2<List<IGUserLite>, bool>> searchUser(String name) async {
    final query = {
      "rank_token" : "${_userId}_${_makeUUID()}",
      "query": name,
      "is_typeahead": true,
      "timezone_offset": DateTime.now().timeZoneOffset.inSeconds
    };
    return (await _apiNet.get("/users/search", queryItems: query, action: IGAPIAction.UserSearch)).tuple2();
  }

  Future<void> _uploadPhotoData(Uint8List imageData, String uploadId, String waterfallId) async {
    final ruploadParams = {};
    ruploadParams["upload_id"] = uploadId;
    ruploadParams["media_type"] = 1;
    ruploadParams["xsharing_user_ids"] = [];
    ruploadParams["image_compression"] = json.encode({
      "quailty": 45,
      "lib_version": "1659.100000",
      "colorspace": "kCGColorSpaceDeviceRGB",
      "lib_name": "uikit"
    });
    
    final stream = Stream.fromIterable(imageData.map((e) => [e]));
    
    (await _apiNet.post("https://i.instagram.com/rupload_igphoto/$uploadId", stream, headers: {
        'Offset' : '0',
        'X-Instagram-Rupload-Params':json.encode(ruploadParams),
        'X-Entity-Name': 'image.jpeg',
        'X-Entity-Type': 'image/jpeg',
        'X-Entity-Length': imageData.length.toString(),
        'media_hash': '',
        'x_fb_photo_waterfall_id': waterfallId,
        HttpHeaders.contentLengthHeader: imageData.length.toString(),
    }, consumes: HTTPBodyConsumes.BLOB, action: IGAPIAction.UploadPhoto)).tuple0();

    return Future.value();
  }

  Future<SXTuple1<IGMedia>> uploadPhotoMedia(Uint8List imageData, {String caption = ""}) async {
    final uploadId = DateTime.now().millisecondsSinceEpoch.toString();
    final waterfallId = _makeUUID(toUpperCase: false, separate: false);

    await _uploadPhotoData(imageData, uploadId, waterfallId);
  
    final timeNow = DateTime.now();
    final body = {
      "lens_model": "unknown",
      "_uuid": _deviceId,
      "_uid": _userId,
      "source_type": "library",
      "scene_capture_type": "standard",
      "edits": {},
      "software": "13.0",
      "geotag_enabled": false,
      "additional_exif_data": {},
      "device_id": _deviceId,
      "scene_type": 1,
      "_csrftoken": _csrftoken,
      "client_timestamp": (timeNow.millisecondsSinceEpoch~/1000).toString(),
      "waterfall_id": waterfallId,
      "allow_multi_configures": false,
      "iso": 800,
      "upload_id": uploadId,
      "camera_position": "unknown",
      "lens_make": "unknown",
      "timezone_offset": timeNow.timeZoneOffset.inSeconds.toString(),
      "disable_comments": false,
      // "date_time_digitized": "2019:07:13 19:10:42",
      "hashtags": "[]",
      // "date_time_original": "2019:07:13 19:10:42",
      "caption": caption
    };
    return (await _apiNet.postSignedJson("/media/configure", body, query: "", action: IGAPIAction.MediaConfig)).tuple1();
  }

  Future<SXTuple1<IGAccountUser>> uploadAvatar(Uint8List imageData) async {
    final uploadId = DateTime.now().millisecondsSinceEpoch.toString();
    final waterfallId = _makeUUID(toUpperCase: false, separate: false);

    final account = (await fetchProfile()).item0;
    
    await _uploadPhotoData(imageData, uploadId, waterfallId);

    final timeNow = DateTime.now();

    String gender = "3";
    switch(account.gender) {
      case IGUserGender.Uncaught:
      case IGUserGender.NotSpecified:
        gender = "3";
        break;
      case IGUserGender.Male:
        gender = "1";
        break;
      case IGUserGender.Female:
        gender = "2";
        break;
    }

    final body = {
      "_csrftoken": _csrftoken,
       "client_timestamp": (timeNow.millisecondsSinceEpoch~/1000).toString(),
      "phone_number": account.phone,
      "timezone_offset": timeNow.timeZoneOffset.inSeconds.toString(),
      "_uuid": _deviceId,
      "_uid": _userId,
      "external_url": account.website,
      "waterfall_id": waterfallId,
      "biography": account.biography,
      "first_name": account.nickname,
      "upload_id": uploadId,
      "username": account.username,
      "device_id": _deviceId,
      "email": account.email,
      "gender": gender,
    };
    return (await _apiNet.postSignedJson("/accounts/edit_profile", body, query: "", action: IGAPIAction.UploadAvatar)).tuple1();
  }
}

final _kIGSystemVolume = Random().nextDouble();

double _calculateLoopCount(int timeNow, int takenAt, double duration) {
  // Policy
  var threshold = 30;

  var inch = Random().nextInt(10);
  if (takenAt % 2 != 0) inch = -inch;
  var clocks = 60 + inch; // [50, 70]
  if (clocks > timeNow - takenAt) {
    clocks = timeNow - takenAt - threshold; //
  }
  if (clocks < 0) clocks = 0;
  double loopCount = clocks / duration;
  if (loopCount >= 5) {
    // < [10, 14]
    loopCount = 5.0 - Random().nextDouble(); // 4.x loops
  }
  // require loop_count ∈ (1, 5)
  return loopCount;
}

Map<String, dynamic> _makeMediaViewMessage(
    IGAPIGateway gateway, IGMedia media) {
  final userId = gateway.profile?.userId;
  final ownerId = media.owner.userId;
  final mediaItem = media.items.first;

  var mediaId = mediaItem.mediaId;
  var duration = mediaItem.duration;
  var audible = mediaItem.audible;
  var videoWidth = mediaItem.originalSize.width;
  var videoHeight = mediaItem.originalSize.height;
  var representationId = mediaItem.videos.first.representationId;
  var trackingToken = media.trackingToken;

  var takenAt = media.takenAt.millisecondsSinceEpoch ~/ 1000;
  var timeNow = DateTime.now().millisecondsSinceEpoch ~/ 1000;

  var loopCount = _calculateLoopCount(timeNow, takenAt, duration);
  if (loopCount < 1) {
    loopCount = 1;
  }

  var playedTimeOnClock = loopCount * duration + Random().nextDouble();
  var timeAsPercent = loopCount % 1;
  var time4VideoPlayedTime = timeNow - Random().nextDouble() * 10 % 4;
  var time4VideoPaused = time4VideoPlayedTime - Random().nextDouble() / 100;
  var time4VideoStartedPlaying = time4VideoPaused - playedTimeOnClock;
  var time4VideoShouldStart =
      time4VideoStartedPlaying - Random().nextDouble() / 10;
  var time4VideoDisplayed = time4VideoShouldStart - Random().nextDouble() / 100;

  var videoPlayedTime = {
    "extra": {
      "a_i": 'organic',
      "a_pk": ownerId,
      "asset_details": 'Looping asset progressive',
      "audio_detected": audible,
      "device_muted": 1,
      "duration": duration,
      "is_dash_eligible": 1,
      "loop_count": loopCount,
      "m_pk": mediaId,
      "m_ts": takenAt,
      "pk": userId,
      "playback_format": 'progressive',
      "played_time_on_clock": playedTimeOnClock,
      "playing_audio": false,
      "radio_type": 'wifi-none',
      "representation_id": representationId,
      "system_volume": _kIGSystemVolume,
      "time": timeAsPercent * duration,
      "timeAsPercent": timeAsPercent,
      "tracking_token": trackingToken,
      "video_height": videoHeight,
      "video_width": videoWidth,
    },
    "module": 'single_feed_profile',
    "name": 'video_played_time',
    "time": time4VideoPlayedTime,
  };

  var videoPaused = {
    "extra": {
      "a_i": 'organic',
      "a_pk": ownerId,
      "asset_details": 'Looping asset progressive',
      "audio_detected": audible,
      "device_muted": 1,
      "duration": duration,
      "is_dash_eligible": 1,
      "loop_count": loopCount,
      "lsp": 0,
      "m_pk": mediaId,
      "m_ts": takenAt,
      "original_start_reason": 'autoplay',
      "pk": userId,
      "playback_format": 'progressive',
      "played_time_on_clock": playedTimeOnClock,
      "playing_audio": false,
      "radio_type": 'wifi-none',
      "reason": 'scroll',
      "representation_id": representationId,
      "system_volume": _kIGSystemVolume,
      "time": timeAsPercent * duration,
      "timeAsPercent": timeAsPercent,
      "tracking_token": trackingToken,
      "video_height": videoHeight,
      "video_width": videoWidth,
    },
    "module": 'single_feed_profile',
    "name": 'video_paused',
    "time": time4VideoPaused,
  };

  var videoStartedPlaying = {
    "extra": {
      "a_i": 'organic',
      "a_pk": ownerId,
      "asset_details": 'Looping asset progressive',
      "audio_detected": audible,
      "device_muted": 1,
      "is_dash_eligible": 1,
      "m_pk": mediaId,
      "m_ts": takenAt,
      "pk": userId,
      "playback_format": 'progressive',
      "playing_audio": false,
      "radio_type": 'wifi-none',
      "reason": 'autoplay',
      "recent_bandwidth": (400 + (Random().nextDouble() * 1000 % 600)).toInt(),
      "representation_id": representationId,
      "system_volume": _kIGSystemVolume,
      "start_delay": 100 + Random().nextDouble(),
      "tracking_token": trackingToken,
      "video_height": videoHeight,
      "video_width": videoWidth,
    },
    "module": 'single_feed_profile',
    "name": 'video_started_playing',
    "time": time4VideoStartedPlaying,
  };

  var videoShouldStart = {
    "extra": {
      "a_i": 'organic',
      "a_pk": ownerId,
      "asset_details": 'Looping asset progressive',
      "audio_detected": audible,
      "device_muted": 1,
      "is_dash_eligible": 1,
      "m_pk": mediaId,
      "m_ts": takenAt,
      "pk": userId,
      "playback_format": 'progressive',
      "radio_type": 'wifi-none',
      "representation_id": representationId,
      "tracking_token": trackingToken,
    },
    "module": 'single_feed_profile',
    "name": 'video_should_start',
    "time": time4VideoShouldStart,
  };

  var videoDisplayed = {
    "extra": {
      "a_i": 'organic',
      "a_pk": ownerId,
      "is_dash_eligible": 1,
      "m_pk": mediaId,
      "m_ts": takenAt,
      "pk": userId,
      "playback_format": 'progressive',
      "radio_type": 'wifi-none',
      "tracking_token": trackingToken,
    },
    "module": 'single_feed_profile',
    "name": 'video_displayed',
    "time": time4VideoDisplayed,
  };

  return {
    "app_id": IGAPIConfig.FACEBOOK_ID,
    "app_ver": "${IGAPIConfig.VERSION}(${IGAPIConfig.BUILD})",
    "channel": "regular",
    "device_id": gateway._deviceId,
    "family_device_id": "",
    "log_type": "client_event",
    "seq": 0,
    "session_id": Uuid().v4().toUpperCase(),
    "time": timeNow,
    "uid": "0",
    "data": [
      videoDisplayed,
      videoShouldStart,
      videoStartedPlaying,
      videoPaused,
      videoPlayedTime
    ]
  };
}
