// ignore_for_file: unnecessary_type_check

import 'dart:async';
import 'dart:io';

import 'package:path_provider/path_provider.dart';
import 'package:sembast/sembast.dart';
import 'package:sembast/sembast_io.dart';

import 'device.dart';

const String QUARK_USER_ID = "@userId";
const String QUARK_DATA_ID = "@dataId";
const String QUARK_AUTH_ID = "@authId";
const String QUARK_TIME_ID = "@timeId";

abstract class QuarkApp {
  // @/flutter_quark/domain/...
  final String domain;

  Database? _usrDB;
  Database? _devDB;
  Database? _abkDB;

  String? _appPath;

  Future<Directory> getAppPath() async {
    if (_appPath == null) {
      Directory rootDir = await getTemporaryDirectory();
      _appPath = [rootDir.path, "flutter_quark/$domain"].join("/");
      return Directory(_appPath!).create(recursive: true);
    }
    return Future.value(Directory(_appPath!));
  }

  Future<void> _init() async {
    await getAppPath();
    
    if (_usrDB == null) {
      _usrDB = await databaseFactoryIo.openDatabase("$_appPath/usr.db");
    }
    if (_devDB == null) {
      _devDB = await databaseFactoryIo.openDatabase("$_appPath/dev.db");
    }
    if (_abkDB == null) {
      _abkDB = await databaseFactoryIo.openDatabase("$_appPath/abk.db");
    }
    Directory("$_appPath/data").createSync(recursive: true);
  }

  Future<Database> _updateOrInsert(
      Database db, String key, Map<String, dynamic> doc) async {
      var store = StoreRef<String, Map<String, dynamic>>.main();
      await store.record(key).put(db, doc);
      // await db.put(doc, key);
      return Future.value(db);
  }

  Future<void> setDevice(String userId, QuarkDevice device) async {
    await _init();
    final doc = device.toJson();
    doc[QUARK_USER_ID] = userId;
    _devDB = await _updateOrInsert(_devDB!, userId, doc);
  }

  Future<QuarkDevice> getDevice(String userId) async {
    await _init();
    var store = StoreRef<String, Map<String, dynamic>>.main();
    final doc = await store.record(userId).get(_devDB!);
    // final doc = await _devDB.get(userId);
    return Future.value(doc is Map ? QuarkDevice.fromJson(doc) : null);
  }

  Future<void> setProfile<T extends QuarkProfile>(
      String userId, T profile) async {
    await _init();
    final doc = profile.toJson();
    _usrDB = await _updateOrInsert(_usrDB!, userId, doc);
  }

  Future<T> getProfileWithAllocator<T extends QuarkProfile>(
      String userId, T fromJson(Map<String, dynamic> doc)) async {
    await _init();
    var store = StoreRef<String, Map<String, dynamic>>.main();
    final doc = await store.record(userId).get(_usrDB!);
    // final doc = await _usrDB.get(userId);
    return Future.value(doc is Map<String, dynamic> ? fromJson(doc) : null);
  }

  Future<List<T>> getAllProfilesWithAllocator<T extends QuarkProfile>(
      T fromJson(Map<String, dynamic> doc)) async {
    await _init();
    var finder = Finder(sortOrders: [SortOrder(QUARK_TIME_ID, false)]);
    var store = StoreRef<String, Map<String, dynamic>>.main();
    var records = await store.find(_usrDB!, finder: finder);
    // final records = await _usrDB.findRecords();
    List<T> list = [];
    for (var record in records) {
      final item = record.value;
      if (item is Map<String, dynamic>) {
        final prof = fromJson(item);
        if (prof is T) list.add(prof);
      }
    }
    return Future.value(list);
  }

  Future<bool> validateDataId(String dataId) async {
    await _init();
    var finder = Finder(filter: Filter.equals(QUARK_DATA_ID, dataId));
    var store = StoreRef<String, Map<String, dynamic>>.main();
    var record = await store.find(_usrDB!, finder: finder);
    // final record = await _usrDB.findRecord(finder);
    return Future.value(record.isNotEmpty);
  }

  Future<void> setAuthPair(String? authId, String userId) async {
    _abkDB = await _updateOrInsert(_abkDB!, authId!,
        {QUARK_AUTH_ID: authId, QUARK_USER_ID: userId});
  }

  // {@"authId":"xxx", @"userId":"yyy"}
  Future<Map<String, String>> getAuthPair(String authId) async {
    await _init();
    var store = StoreRef<String, Map<String, dynamic>>.main();
    final doc = await store.record(authId).get(_abkDB!);
    // final doc = await _abkDB.get(authId);
    return Future.value(doc is Map<String, String> ? doc : null);
  }

  Future<Directory> getDataDirectory(String dataId) async {
    await _init();
    final path = [_appPath, "data/$dataId"].join("/");
    return Directory(path).create(recursive: true);
  }

  Future<void> cleanInvalidData() async {
    final dd = await this.getAppPath();
    final ddr = Directory([dd.path, "data"].join("/"));
    if (ddr.existsSync() != true) return;
    final subs = ddr.listSync(followLinks: false);
    for (var sub in subs) {
      if (sub is! Directory) continue;
      String dataId = sub.path.split("/").last;

      bool valid = await this.validateDataId(dataId);
      if (valid != true) {
        sub.deleteSync(recursive: true);
      }
    }
  }

  // domain is prefix 8 in hmacsha256("$name-v$version", "quark");
  QuarkApp(this.domain);
}

abstract class QuarkProfile {
  String get userId;

  String get dataId;

  DateTime get loggedInTime; // 登录时间TS13

  /// 是否过期 暗示是否需要登录
  bool get expired => loggedInTime.millisecondsSinceEpoch == 0;

  Map<String, dynamic> toJson();

  QuarkProfile(Map<String, dynamic> json) {
    assert(json[QUARK_USER_ID] is String);
    assert(json[QUARK_DATA_ID] is String);
    assert(json[QUARK_TIME_ID] is int);
  }
}

