
// ignore_for_file: unnecessary_type_check

import 'dart:io';

import 'package:cookie_jar/cookie_jar.dart';
import 'package:igapi/src/network.dart';
import 'package:uuid/uuid.dart';

import 'device.dart';
import 'models.dart';
import 'quark.dart';

class _IGPlatform extends QuarkApp {
  // hmacsha256("instagram-v1", "quark");
  _IGPlatform() : super("0756f0cb");

  Future<IGProfile?> getProfile(String userId) async {
    return super.getProfileWithAllocator(userId, (doc)=> IGProfile.fromJson(doc));
  }

  Future<List<IGProfile>> getAllProfiles() async {
    return super.getAllProfilesWithAllocator((doc)=> IGProfile.fromJson(doc));
  }
}

final _IGPlatform _igapp = _IGPlatform();

class IGProfile extends QuarkProfile {
  Map<String, dynamic> _obj = {};

  @override
  Map<String, dynamic> toJson() {
    return Map.from(_obj);
  }

  factory IGProfile.fromJson(Map<String, dynamic> json) {
    return IGProfile(json);
  }

  IGProfile(this._obj) : super(_obj);

  // 查看本地
  static Future<List<IGProfile>> getAllPersistProfiles() async {
    final items = await _igapp.getAllProfiles();
    return Future.value(items);
  }

  @override
  String get dataId => _obj[QUARK_DATA_ID];

  @override
  String get userId => _obj[QUARK_USER_ID];

  @override
  DateTime get loggedInTime => DateTime.fromMillisecondsSinceEpoch(_obj[QUARK_TIME_ID]);

  /// 用户名
  String get username => _obj["username"];

  /// 昵称
  String get nickname => _obj["nickname"];

  /// 头像(缩略图)
  String get thumbnailAvatarURL => _obj["thumbnailAvatarURL"];

  /// 是否使用了匿名头像
  bool get anonymousAvatar => _obj["anonymousAvatar"];

  /// 是否为私有用户
  bool get private => _obj["private"];

  /// 是否为认证用户
  bool get verified => _obj["verified"];
}

class IGState {  
  // Auth
  String? authId;
  int loginAttemptCount = 0;
  IGChallengeLock? challengeLock;
  IGTwoFactorLock? twoFactorLock;

  void _resetAuth() {
    authId = null;
    loginAttemptCount = 0;
    challengeLock = null;
    twoFactorLock = null;
  }
}

class IGAccount {
  final IGState state = IGState();
  String? _dataId;
  IGProfile? _profile;
  QuarkDevice _device;
  CookieJar _cookieJar;
  APINet?    _apiNet;

   void _reset(IGAccount account) {
     state._resetAuth();
    if (this == account) return;
    _dataId = account.dataId;
    _profile = account.profile;
    _device = account.device;
    _cookieJar = account.cookieJar;
    _apiNet = account.apiNet;
  }

  IGProfile? get profile => _profile;
  QuarkDevice get device => _device;
  CookieJar get cookieJar => _cookieJar;
  APINet? get apiNet => _apiNet;

  String? get dataId {
    return profile?.dataId ?? _dataId;
  }

  static Future<PersistCookieJar> _makeCookieJar(String dataId) async {
    final dd = await _igapp.getDataDirectory(dataId);
    final dir = Directory("${dd.path}/cookies");
    dir.createSync(recursive: true);
    return Future.value(PersistCookieJar(ignoreExpires: true, storage: FileStorage(dir.path)));
  }

  static final Set<IGAccount> _persistCache = Set();
  static final Map<String, IGAccount> _ephemeralCache = {};

  IGAccount(this._profile, this._device, this._cookieJar) {
    if (profile != null) {
      _persistCache.add(this);
    }
  }

  void offline() {
    _persistCache.remove(this);
  }

  /// 同族Account共享数据
  void reloadFamily(IGAccount account) {
    final userId = account.profile?.userId;
    for (var _acc in _persistCache) {
      if (_acc.profile?.userId != userId) continue;
      _acc._reset(account);
    }
    _persistCache.add(account);
  }

  /// 持久化账户认证凭据失效
  Future<void> onAuthExpired() async {
    assert(profile != null);
    profile!._obj[QUARK_TIME_ID] = 0;
    await _igapp.setProfile(profile!.userId, profile!);
    reloadFamily(this);
    return Future.value();
  }

  /// 新增持久化账户(可能会覆盖本地已有的持久化账户)
  Future<void> onLoggedIn(IGLoggedInUser? user, [skipIfNull = false]) async {
    if (skipIfNull == true && user == null) return Future.value();

    assert(user != null);
    
    final json = user!.toJson();

    final userId = user.userId;
    
    json[QUARK_USER_ID] = userId;
    json[QUARK_DATA_ID] = dataId;
    json[QUARK_TIME_ID] = DateTime.now().millisecondsSinceEpoch;

    _profile = IGProfile.fromJson(json);
    await _igapp.setProfile(userId, profile!);

    await _igapp.setDevice(userId, device);

    await _igapp.setAuthPair(state.authId, userId);
    if (user.username != null) {
      await _igapp.setAuthPair("0:${user.username}", userId);
    }
    if (user.phone != null) {
      await _igapp.setAuthPair("0:${user.phone}", userId);
      if (user.phone!.startsWith("+")) {
        await _igapp.setAuthPair("0:${user.phone!.substring(1)}", userId);
      }
    }
    if (user.fbuid != null) {
      await _igapp.setAuthPair("1:${user.fbuid}", userId);
    }
    // auth email from IGAccountUser

    await _igapp.cleanInvalidData();

    state._resetAuth();

    _ephemeralCache.clear();

    reloadFamily(this);

    return Future.value();
  }

  static Future<IGAccount?> getPersist(String? userId) async {
    if (userId is! String) return Future.value(null);

    final prof = await _igapp.getProfile(userId);
    if (prof is! IGProfile || prof.dataId is! String) return Future.value(null);

    var device = await _igapp.getDevice(userId);
    if (device is! QuarkDevice) {
      device = QuarkDevice.random();
    }

    final cookieJar = await _makeCookieJar(prof.dataId);

    final account = IGAccount(prof, device, cookieJar);
    // account._dataId = prof.dataId
    account._apiNet = APINet(account);

    return Future.value(account);
  } 

  // username, email, phone, fbuid etc.
  static Future<IGAccount> forLogin(String username, {bool facebook = false, bool reuse = true}) async {
    IGAccount? account;

    final String authId = "${facebook != true ? 0 : 1}:$username";
    final match = _ephemeralCache[authId];
    if (match != null) return Future.value(match);
    
    if (reuse == true) {
      final pair = await _igapp.getAuthPair(authId);
      if (pair is Map) {
        account = await getPersist(pair[QUARK_USER_ID]);
      }
    }
    
    if (account is! IGAccount) {
      final dataId = Uuid().v4().replaceAll("-", "");
      final cookieJar = await _makeCookieJar(dataId);
      account = IGAccount(null, QuarkDevice.random(), cookieJar);
      account._dataId = dataId;
      account._apiNet = APINet(account);
      _ephemeralCache[authId] = account;
    }

    account.state.authId = authId;

    return Future.value(account);
  }
}
